@@ -33,6 +33,10 @@
 #include <QMovie>
 #include <QPainter>
 #include <qstandardpaths.h>
+#ifdef Q_OS_MACOS
+#include <qfontdatabase.h>
+#include <QtMac>
+#endif
 
 #include <kwindowinfo.h>
 #include <kwindowsystem.h>
@@ -70,6 +74,9 @@ KStatusNotifierItem::~KStatusNotifierItem()
     if (!qApp->closingDown()) {
         delete d->menu;
     }
+    if (d->associatedWidget) {
+        KWindowSystem::self()->disconnect(d->associatedWidget);
+    }
     delete d;
 }
 
@@ -251,7 +258,7 @@ void KStatusNotifierItem::setAttentionMovieByName(const QString &name)
     d->movieName = name;
 
     delete d->movie;
-    d->movie = 0;
+    d->movie = nullptr;
 
     emit d->statusNotifierItemDBus->NewAttentionIcon();
 
@@ -268,6 +275,36 @@ QString KStatusNotifierItem::attentionMovieName() const
 
 //ToolTip
 
+#ifdef Q_OS_MACOS
+static void setTrayToolTip(KStatusNotifierLegacyIcon *systemTrayIcon, const QString &title, const QString &subTitle)
+{
+    if (systemTrayIcon) {
+        bool tEmpty = title.isEmpty(),
+                    stEmpty = subTitle.isEmpty();
+                    if (tEmpty) {
+                        if (!stEmpty) {
+                            systemTrayIcon->setToolTip(subTitle);
+                        } else {
+                            systemTrayIcon->setToolTip(title);
+                        }
+                    } else {
+                        if (stEmpty) {
+                            systemTrayIcon->setToolTip(title);
+                        } else {
+                            systemTrayIcon->setToolTip(title + QStringLiteral("\n") + subTitle);
+                        }
+                    }
+    }
+}
+#else
+static void setTrayToolTip(KStatusNotifierLegacyIcon *systemTrayIcon, const QString &title, const QString &)
+{
+    if (systemTrayIcon) {
+        systemTrayIcon->setToolTip(title);
+    }
+}
+#endif
+
 void KStatusNotifierItem::setToolTip(const QString &iconName, const QString &title, const QString &subTitle)
 {
     if (d->toolTipIconName == iconName &&
@@ -280,10 +317,10 @@ void KStatusNotifierItem::setToolTip(const QString &iconName, const QString &tit
     d->toolTipIconName = iconName;
 
     d->toolTipTitle = title;
-    if (d->systemTrayIcon) {
-        d->systemTrayIcon->setToolTip(title);
-    }
-
+//     if (d->systemTrayIcon) {
+//         d->systemTrayIcon->setToolTip(title);
+//     }
+    setTrayToolTip(d->systemTrayIcon, title, subTitle);
     d->toolTipSubTitle = subTitle;
     emit d->statusNotifierItemDBus->NewToolTip();
 }
@@ -301,9 +338,10 @@ void KStatusNotifierItem::setToolTip(const QIcon &icon, const QString &title, co
     d->toolTipIcon = icon;
 
     d->toolTipTitle = title;
-    if (d->systemTrayIcon) {
-        d->systemTrayIcon->setToolTip(title);
-    }
+//     if (d->systemTrayIcon) {
+//         d->systemTrayIcon->setToolTip(title);
+//     }
+    setTrayToolTip(d->systemTrayIcon, title, subTitle);
 
     d->toolTipSubTitle = subTitle;
     emit d->statusNotifierItemDBus->NewToolTip();
@@ -350,9 +388,10 @@ void KStatusNotifierItem::setToolTipTitle(const QString &title)
 
     d->toolTipTitle = title;
     emit d->statusNotifierItemDBus->NewToolTip();
-    if (d->systemTrayIcon) {
-        d->systemTrayIcon->setToolTip(title);
-    }
+//     if (d->systemTrayIcon) {
+//         d->systemTrayIcon->setToolTip(title);
+//     }
+    setTrayToolTip(d->systemTrayIcon, title, d->toolTipSubTitle);
 }
 
 QString KStatusNotifierItem::toolTipTitle() const
@@ -367,6 +406,9 @@ void KStatusNotifierItem::setToolTipSubTitle(const QString &subTitle)
     }
 
     d->toolTipSubTitle = subTitle;
+#ifdef Q_OS_MACOS
+    setTrayToolTip(d->systemTrayIcon, d->toolTipTitle, subTitle);
+#endif
     emit d->statusNotifierItemDBus->NewToolTip();
 }
 
@@ -383,7 +425,7 @@ void KStatusNotifierItem::setContextMenu(QMenu *menu)
     }
 
     if (!menu) {
-        d->menu = 0;
+        d->menu = nullptr;
         return;
     }
 
@@ -408,7 +450,7 @@ void KStatusNotifierItem::setContextMenu(QMenu *menu)
     }
 
     d->menu = menu;
-    d->menu->setParent(0);
+    d->menu->setParent(nullptr);
 }
 
 QMenu *KStatusNotifierItem::contextMenu() const
@@ -420,13 +462,31 @@ void KStatusNotifierItem::setAssociatedWidget(QWidget *associatedWidget)
 {
     if (associatedWidget) {
         d->associatedWidget = associatedWidget->window();
-    } else {
-        d->associatedWidget = 0;
+        d->associatedWidgetPos = QPoint(-1, -1);
+
+        QObject::connect(KWindowSystem::self(), &KWindowSystem::windowAdded,
+            d->associatedWidget, [this](WId id) {
+                if(d->associatedWidget->winId() == id) {
+                    d->associatedWidget->move(d->associatedWidgetPos);
+                }
+            }
+        );
+
+        QObject::connect(KWindowSystem::self(), &KWindowSystem::windowRemoved,
+            d->associatedWidget, [this](WId id) {
+                if(d->associatedWidget->winId() == id) {
+                    d->associatedWidgetPos = d->associatedWidget->pos();
+                }
+            }
+        );
+    } else if (d->associatedWidget) {
+        KWindowSystem::self()->disconnect(d->associatedWidget);
+        d->associatedWidget = nullptr;
     }
 
     if (d->systemTrayIcon) {
         delete d->systemTrayIcon;
-        d->systemTrayIcon = 0;
+        d->systemTrayIcon = nullptr;
         d->setLegacySystemTrayEnabled(true);
     }
 
@@ -440,7 +500,7 @@ void KStatusNotifierItem::setAssociatedWidget(QWidget *associatedWidget)
             connect(action, SIGNAL(triggered(bool)), this, SLOT(minimizeRestore()));
         }
 
-        KWindowInfo info = KWindowSystem::windowInfo(d->associatedWidget->winId(), NET::WMDesktop);
+        KWindowInfo info(d->associatedWidget->winId(), NET::WMDesktop);
         d->onAllDesktops = info.onAllDesktops();
     } else {
         if (d->menu && d->hasQuit) {
@@ -515,7 +575,15 @@ void KStatusNotifierItem::showMessage(const QString &title, const QString &messa
     }
 
     uint id = 0;
-    d->notificationsClient->Notify(d->title, id, icon, title, message, QStringList(), QVariantMap(), timeout);
+#ifdef Q_OS_MACOS
+    if (d->systemTrayIcon) {
+        // Growl is not needed anymore for QSystemTrayIcon::showMessage() since OS X 10.8
+        d->systemTrayIcon->showMessage(title, message, QSystemTrayIcon::Information, timeout);
+    } else
+#endif
+    {
+        d->notificationsClient->Notify(d->title, id, icon, title, message, QStringList(), QVariantMap(), timeout);
+    }
 }
 
 QString KStatusNotifierItem::title() const
@@ -529,6 +597,9 @@ void KStatusNotifierItem::activate(const QPoint &pos)
     //FIXME: always true?
     if (d->status == NeedsAttention) {
         d->status = Active;
+#ifdef Q_OS_MACOS
+        QtMac::setBadgeLabelText(QString());
+#endif
         emit d->statusNotifierItemDBus->NewStatus(QString::fromLatin1(metaObject()->enumerator(metaObject()->indexOfEnumerator("ItemStatus")).valueToKey(d->status)));
     }
 
@@ -566,7 +637,7 @@ bool KStatusNotifierItemPrivate::checkVisibility(QPoint pos, bool perform)
     }
 #endif
 #else
-    KWindowInfo info1 = KWindowSystem::windowInfo(associatedWidget->winId(), NET::XAWMState | NET::WMState | NET::WMDesktop);
+    KWindowInfo info1(associatedWidget->winId(), NET::XAWMState | NET::WMState | NET::WMDesktop);
     // mapped = visible (but possibly obscured)
     bool mapped = (info1.mappingState() == NET::Visible) && !info1.isMinimized();
 
@@ -591,8 +662,7 @@ bool KStatusNotifierItemPrivate::checkVisibility(QPoint pos, bool perform)
                 break;
             }
 
-            KWindowInfo info2 = KWindowSystem::windowInfo(id,
-                                NET::WMDesktop | NET::WMGeometry | NET::XAWMState | NET::WMState | NET::WMWindowType);
+            KWindowInfo info2(id, NET::WMDesktop | NET::WMGeometry | NET::XAWMState | NET::WMState | NET::WMWindowType);
 
             if (info2.mappingState() != NET::Visible) {
                 continue; // not visible on current desktop -> ignore
@@ -647,7 +717,7 @@ bool KStatusNotifierItemPrivate::checkVisibility(QPoint pos, bool perform)
 
 bool KStatusNotifierItem::eventFilter(QObject *watched, QEvent *event)
 {
-    if (d->systemTrayIcon == 0) {
+    if (d->systemTrayIcon == nullptr) {
         //FIXME: ugly ugly workaround to weird QMenu's focus problems
         if (watched == d->menu &&
                 (event->type() == QEvent::WindowDeactivate || (event->type() == QEvent::MouseButtonRelease && static_cast<QMouseEvent *>(event)->button() == Qt::LeftButton))) {
@@ -666,12 +736,13 @@ KStatusNotifierItemPrivate::KStatusNotifierItemPrivate(KStatusNotifierItem *item
     : q(item),
       category(KStatusNotifierItem::ApplicationStatus),
       status(KStatusNotifierItem::Passive),
-      movie(0),
-      menu(0),
-      titleAction(0),
-      statusNotifierWatcher(0),
-      notificationsClient(0),
-      systemTrayIcon(0),
+      movie(nullptr),
+      menu(nullptr),
+      associatedWidget(nullptr),
+      titleAction(nullptr),
+      statusNotifierWatcher(nullptr),
+      notificationsClient(nullptr),
+      systemTrayIcon(nullptr),
       hasQuit(false),
       onAllDesktops(false),
       standardActionsEnabled(true)
@@ -701,8 +772,24 @@ void KStatusNotifierItemPrivate::init(const QString &extraId)
     if (title.isEmpty()) {
         title = QCoreApplication::applicationName();
     }
+#ifdef Q_OS_MACOS
+    // OS X doesn't have texted separators so we emulate QAction::addSection():
+    // we first add an action with the desired text (title) and icon
+    titleAction = m->addAction(qApp->windowIcon(), title);
+    // this action should be disabled
+    titleAction->setEnabled(false);
+    // Give the titleAction a visible menu icon:
+    // Systray icon and menu ("menu extra") are often used by applications that provide no other interface.
+    // It is thus reasonable to show the application icon in the menu; Finder, Dock and App Switcher
+    // all show it in addition to the application name (and Apple's input "menu extra" also shows icons).
+    titleAction->setIconVisibleInMenu(true);
+    m->addAction(titleAction);
+    // now add a regular separator
+    m->addSeparator();
+#else
     titleAction = m->addSection(qApp->windowIcon(), title);
     m->setTitle(title);
+#endif
     q->setContextMenu(m);
 
     QAction *action = new QAction(q);
@@ -777,7 +864,7 @@ void KStatusNotifierItemPrivate::serviceChange(const QString &name, const QStrin
         qCDebug(LOG_KNOTIFICATIONS) << "Connection to the KStatusNotifierWatcher lost";
         setLegacyMode(true);
         delete statusNotifierWatcher;
-        statusNotifierWatcher = 0;
+        statusNotifierWatcher = nullptr;
     } else if (oldOwner.isEmpty()) {
         //registered
         setLegacyMode(false);
@@ -817,7 +904,7 @@ void KStatusNotifierItemPrivate::legacyActivated(QSystemTrayIcon::ActivationReas
 
 void KStatusNotifierItemPrivate::setLegacySystemTrayEnabled(bool enabled)
 {
-    if (enabled == (systemTrayIcon != 0)) {
+    if (enabled == (systemTrayIcon != nullptr)) {
         // already in the correct state
         return;
     }
@@ -848,7 +935,7 @@ void KStatusNotifierItemPrivate::setLegacySystemTrayEnabled(bool enabled)
         }
     } else {
         delete systemTrayIcon;
-        systemTrayIcon = 0;
+        systemTrayIcon = nullptr;
 
         if (menu) {
             menu->setWindowFlags(Qt::Window);
@@ -857,7 +944,7 @@ void KStatusNotifierItemPrivate::setLegacySystemTrayEnabled(bool enabled)
 
     if (menu) {
         QMenu *m = menu;
-        menu = 0;
+        menu = nullptr;
         q->setContextMenu(m);
     }
 }
@@ -865,22 +952,72 @@ void KStatusNotifierItemPrivate::setLegacySystemTrayEnabled(bool enabled)
 void KStatusNotifierItemPrivate::syncLegacySystemTrayIcon()
 {
     if (status == KStatusNotifierItem::NeedsAttention) {
-        if (!movieName.isNull()) {
-            if (!movie) {
-                movie = new QMovie(movieName);
+#ifdef Q_OS_MACOS
+        QtMac::setBadgeLabelText(QString(QChar(0x26a0))/*QStringLiteral("!")*/);
+        if (attentionIconName.isNull() && attentionIcon.isNull()) {
+            // code adapted from kmail's KMSystemTray::updateCount()
+            int overlaySize = 22;
+            QIcon attnIcon = qApp->windowIcon();
+            if (!attnIcon.availableSizes().isEmpty()) {
+                overlaySize = attnIcon.availableSizes().at(0).width();
+            }
+            QFont labelFont = QFontDatabase::systemFont(QFontDatabase::GeneralFont);
+            labelFont.setBold(true);
+            QFontMetrics qfm(labelFont);
+            float attnHeight = overlaySize * 0.667;
+            if (qfm.height() > attnHeight) {
+                float labelSize = attnHeight;
+                labelFont.setPointSizeF(labelSize);
+            }
+            // Paint the label in a pixmap
+            QPixmap overlayPixmap(overlaySize, overlaySize);
+            overlayPixmap.fill(Qt::transparent);
+
+            QPainter p(&overlayPixmap);
+            p.setFont(labelFont);
+            p.setBrush(Qt::NoBrush);
+            // this sort of badge/label is red on OS X
+            p.setPen(QColor(224,0,0));
+            p.setOpacity(1.0);
+            // use U+2022, the Unicode bullet
+            p.drawText(overlayPixmap.rect(), Qt::AlignRight|Qt::AlignTop, QString(QChar(0x2022)));
+            p.end();
+
+            QPixmap iconPixmap = attnIcon.pixmap(overlaySize, overlaySize);
+            QPainter pp(&iconPixmap);
+            pp.drawPixmap(0, 0, overlayPixmap);
+            pp.end();
+            systemTrayIcon->setIcon(iconPixmap);
+        } else
+#endif
+        {
+            if (!movieName.isNull()) {
+                if (!movie) {
+                    movie = new QMovie(movieName);
+                }
+                systemTrayIcon->setMovie(movie);
+            } else if (!attentionIconName.isNull()) {
+                systemTrayIcon->setIcon(QIcon::fromTheme(attentionIconName));
+            } else {
+                systemTrayIcon->setIcon(attentionIcon);
             }
-            systemTrayIcon->setMovie(movie);
-        } else if (!attentionIconName.isNull()) {
-            systemTrayIcon->setIcon(QIcon::fromTheme(attentionIconName));
-        } else {
-            systemTrayIcon->setIcon(attentionIcon);
         }
     } else {
+#ifdef Q_OS_MACOS
+        if (!iconName.isNull()) {
+            QIcon theIcon = QIcon::fromTheme(iconName);
+            systemTrayIcon->setIconWithMask(theIcon, status==KStatusNotifierItem::Passive);
+        } else {
+            systemTrayIcon->setIconWithMask(icon, status==KStatusNotifierItem::Passive);
+        }
+        QtMac::setBadgeLabelText(QString());
+#else
         if (!iconName.isNull()) {
             systemTrayIcon->setIcon(QIcon::fromTheme(iconName));
         } else {
             systemTrayIcon->setIcon(icon);
         }
+#endif
     }
 
     systemTrayIcon->setToolTip(toolTipTitle);
@@ -947,7 +1084,7 @@ void KStatusNotifierItemPrivate::hideMenu()
 
 void KStatusNotifierItemPrivate::minimizeRestore(bool show)
 {
-    KWindowInfo info = KWindowSystem::windowInfo(associatedWidget->winId(), NET::WMDesktop | NET::WMFrameExtents);
+    KWindowInfo info(associatedWidget->winId(), NET::WMDesktop);
     if (show) {
         if (onAllDesktops) {
             KWindowSystem::setOnAllDesktops(associatedWidget->winId(), true);
@@ -955,11 +1092,10 @@ void KStatusNotifierItemPrivate::minimizeRestore(bool show)
             KWindowSystem::setCurrentDesktop(info.desktop());
         }
 
-        associatedWidget->move(info.frameGeometry().topLeft()); // avoid placement policies
+        auto state = associatedWidget->windowState() & ~Qt::WindowMinimized;
+        associatedWidget->setWindowState(state);
         associatedWidget->show();
         associatedWidget->raise();
-        KWindowSystem::raiseWindow(associatedWidget->winId());
-        KWindowSystem::forceActiveWindow(associatedWidget->winId());
     } else {
         onAllDesktops = info.onAllDesktops();
         associatedWidget->hide();
